/*
    Pacman arcade machine emulator

    Copyright (c) 1997-2003,2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "arcade.h"

#include <string.h>

// Namco 3-channel Wave Sound Generator wave data (8 waveforms with 32 4-bit entries each)
static unsigned char NamcoSoundPROM[] = {
	0x07,0x09,0x0A,0x0B,0x0C,0x0D,0x0D,0x0E,0x0E,0x0E,0x0D,0x0D,0x0C,0x0B,0x0A,0x09,
	0x07,0x05,0x04,0x03,0x02,0x01,0x01,0x00,0x00,0x00,0x01,0x01,0x02,0x03,0x04,0x05,
	0x07,0x0C,0x0E,0x0E,0x0D,0x0B,0x09,0x0A,0x0B,0x0B,0x0A,0x09,0x06,0x04,0x03,0x05,
	0x07,0x09,0x0B,0x0A,0x08,0x05,0x04,0x03,0x03,0x04,0x05,0x03,0x01,0x00,0x00,0x02,
	0x07,0x0A,0x0C,0x0D,0x0E,0x0D,0x0C,0x0A,0x07,0x04,0x02,0x01,0x00,0x01,0x02,0x04,
	0x07,0x0B,0x0D,0x0E,0x0D,0x0B,0x07,0x03,0x01,0x00,0x01,0x03,0x07,0x0E,0x07,0x00,
	0x07,0x0D,0x0B,0x08,0x0B,0x0D,0x09,0x06,0x0B,0x0E,0x0C,0x07,0x09,0x0A,0x06,0x02,
	0x07,0x0C,0x08,0x04,0x05,0x07,0x02,0x00,0x03,0x08,0x05,0x01,0x03,0x06,0x03,0x01,
	0x00,0x08,0x0F,0x07,0x01,0x08,0x0E,0x07,0x02,0x08,0x0D,0x07,0x03,0x08,0x0C,0x07,
	0x04,0x08,0x0B,0x07,0x05,0x08,0x0A,0x07,0x06,0x08,0x09,0x07,0x07,0x08,0x08,0x07,
	0x07,0x08,0x06,0x09,0x05,0x0A,0x04,0x0B,0x03,0x0C,0x02,0x0D,0x01,0x0E,0x00,0x0F,
	0x00,0x0F,0x01,0x0E,0x02,0x0D,0x03,0x0C,0x04,0x0B,0x05,0x0A,0x06,0x09,0x07,0x08,
	0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,
	0x0F,0x0E,0x0D,0x0C,0x0B,0x0A,0x09,0x08,0x07,0x06,0x05,0x04,0x03,0x02,0x01,0x00,
	0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F,
	0x00,0x01,0x02,0x03,0x04,0x05,0x06,0x07,0x08,0x09,0x0A,0x0B,0x0C,0x0D,0x0E,0x0F
};

PacmanMachine::PacmanMachine()
{
    // Initialize the CPU and the RAM
    cpu_ = new Z80( *this );
    memset( ram_, 0xFF, sizeof(ram_) );

    // Set the sound PROM to the default values for the original Namco chip
    setSoundPROM( NamcoSoundPROM );

    // Set the DIP switches to a default setting
    dip_switches_ = DipPlay_OneCoinOneGame | DipBonus_10000 | DipLives_3 | DipDifficulty_Normal | DipGhostNames_Normal;

    // Reset the machine
    reset();

    // Initialize the video character translation tables: video memory has a very
    // peculiar arrangement in Pacman so we precompute a few tables to move around faster
    for( int i=0x000; i<0x400; i++ ) {
        int x, y;

        if( i < 0x040 ) {
          x = 29 - (i & 0x1F);
          y = 34 + (i >> 5);
        }
        else if( i >= 0x3C0 ) {
          x = 29 - (i & 0x1F);
          y = ((i-0x3C0) >> 5);
        }
        else {
          x = 27 - ((i-0x40) >> 5);
          y = 2 + ((i-0x40) & 0x1F);
        }
        vchar_to_x_[i] = x;
        vchar_to_y_[i] = y;
        if( (y >= 0) && (y < 36) && (x >= 0) && (x < 28) )
            vchar_to_i_[i] = y*28 + x;
        else
            vchar_to_i_[i] = 0x3FF;
    }
}

PacmanMachine::~PacmanMachine()
{
    delete cpu_;
}

/*
    For Z80 Environment: read a byte from memory.
*/
unsigned char PacmanMachine::readByte( unsigned addr ) 
{
    addr &= 0x7FFF;

    if( addr < sizeof(ram_) )
        return ram_[addr];

    // Address is not in RAM, check to see if it's a memory mapped register
    switch( addr & 0xFFC0 ) {
    // IN0
    case 0x5000: 
        // bit 0 : up
        // bit 1 : left
        // bit 2 : right
        // bit 3 : down
        // bit 4 : rack test -> 0x10=off, 0=on
        // bit 5 : coin 1
        // bit 6 : coin 2
        return port1_;
    // IN1
    case 0x5040: 
        // bit 0 : up
        // bit 1 : left
        // bit 2 : right
        // bit 3 : down
        // bit 5 : start 1
        // bit 6 : start 2
        // bit 7 : cabinet -> 0x80=upright, 0x00=cocktail
        return (port2_ & 0xF0) | (port1_ & 0x0F);
    // DSW1
    case 0x5080:
        // bits 0,1 : coinage -> 0=free play, 1=1 coin/play, 2=1 coin/2 play, 3=2 coin/1 play
        // bits 2,3 : lives -> 0x0=1, 0x4=2, 0x8=3, 0xC=5
        // bits 4,5 : bonus life -> 0=10000, 0x10=15000, 0x20=20000, 0x30=none
        // bit  6   : difficulty -> 0x40=normal, 0=hard
        // bit  7   : ghost name -> 0x80=normal, 0=alternate
        return dip_switches_;
    // Watchdog reset
    case 0x50C0:
        return 0xFF;
    }

    return 0xFF;
}

/*
    For Z80 Environment: write a byte to memory.
*/
void PacmanMachine::writeByte( unsigned addr, unsigned char b )
{
    addr &= 0x7FFF;

    if( addr < 0x4000 ) {
        // This is a ROM address, do not write into it!
    }
    else if( addr < 0x4400 ) {
        // Video memory
        ram_[addr] = b;
        video_mem_[ vchar_to_i_[addr-0x4000] ] = b;
    }
    else if( addr < 0x4800 ) {
        // Color memory
        ram_[addr] = b;
        color_mem_[ vchar_to_i_[addr-0x4400] ] = b;
    }
    else if( addr < 0x4FF0 ) {
        // Standard memory
        ram_[addr] = b;
    }
    else if( addr < 0x5000 ) {
        // Sprites
        ram_[addr] = b;

        unsigned idx = (addr - 0x4FF0) / 2;

        if( addr & 1 ) {
            sprites_[ idx ].color = b;
        }
        else {
            sprites_[ idx ].n = b >> 2;
            sprites_[ idx ].mode = b & 0x03;
        }
    }
    else {
        // Memory mapped ports
        switch( addr ) {
        case 0x5000:
            interrupt_enabled_ = b;
            break;
        case 0x5001:
            // Sound enable
            break;
        case 0x5003:
            // Flip screen
            break;
        case 0x5004:
            // Player 1 start lamp
            break;
        case 0x5005:
            // Player 2 start lamp
            break;
        case 0x5006:
            // Coin lockout
            break;
        case 0x5007:
            // Coin counter
            break;
        case 0x50c0:
            // Watchdog reset
            break;
        default:
            if( addr >= 0x5040 && addr < 0x5060 ) {
                // Sound registers
                sound_regs_[ addr-0x5040 ] = b;
            }
            else if( addr >= 0x5060 && addr < 0x5070 ) {
                // Sprite coordinates, x/y pairs for 8 sprites
                unsigned idx = (addr-0x5060) / 2;

                if( addr & 1 ) {
                    sprites_[ idx ].y = 272 - b + 1;
                }
                else {
                    sprites_[ idx ].x = 240 - b - 1;

                    if( idx <= 2 ) {
                        // In Pacman the first ew sprites must be further offset 
                        // to the left to get a correct display (is this a hack?)
                        sprites_[ idx ].x -= 1;
                    }
                }
            }
        }
    }
}

/*
    For Z80 Environment: read from a port.

    Note: all ports in Pacman are memory mapped so they are read with readByte().
*/
unsigned char PacmanMachine::readPort( unsigned port ) 
{
    return 0;
}

/*
    For Z80 Environment: write to a port.
*/
void PacmanMachine::writePort( unsigned addr, unsigned char b )
{
    if( addr == 0 ) {
        // Sets the interrupt mode for the next CPU interrupt
        interrupt_ = b;
    }
}

void PacmanMachine::setROM( const unsigned char * rom )
{
    memcpy( ram_, rom, 0x4000 );
}

void PacmanMachine::setVideoROMs( const unsigned char * charset, const unsigned char * spriteset )
{
    memcpy( charset_rom_, charset, sizeof(charset_rom_) );
    memcpy( spriteset_rom_, spriteset, sizeof(spriteset_rom_) );

    decodeCharSet( charset_rom_, charmap_ );
    decodeSprites( spriteset_rom_, spritemap_ );
}

void PacmanMachine::setColorROMs( const unsigned char * palette, const unsigned char * color )
{
    unsigned decoded_palette[0x20];

    memcpy( palette_data_, palette, sizeof(palette_data_) );
    memcpy( color_data_, color, sizeof(color_data_) );

    int i;

    for( i=0x00; i<0x20; i++ ) {
        decoded_palette[i] = decodePaletteByte( palette[i] );
    }

    for( i=0; i<256; i++ ) {
        palette_[i] = decoded_palette[ color[i] & 0x0F ];
    }
}

void PacmanMachine::setSoundPROM( const unsigned char * prom )
{
    memcpy( sound_prom_, prom, sizeof(sound_prom_) );

    for( int i=0; i<32*8; i++ ) {
        sound_wave_data_[i] = (int)*prom++ - 8;
    }
}

void PacmanMachine::reset()
{
    cpu_->reset();
    interrupt_enabled_ = 0;
    interrupt_ = 0;
    port1_ = 0xFF;
    port2_ = 0xFF;
    memset( ram_+0x4000, 0, 0x1000 );
    memset( color_mem_, 0, sizeof(color_mem_) );
    memset( video_mem_, 0, sizeof(video_mem_) );
}

/*
    Run the machine for one frame.
*/
int PacmanMachine::run()
{
    // Run until the CPU has executed the number of cycles per frame
    while( cpu_->getCycles() < CpuCyclesPerFrame ) {
        cpu_->step();
    }

    // Update the cycle count in preparation for the next frame
    cpu_->setCycles( cpu_->getCycles() - CpuCyclesPerFrame );

    // If interrupts are enabled, force a CPU interrupt with the mode
    // set by the program
    if( interrupt_enabled_ ) {
        cpu_->interrupt( interrupt_ );
    }

    // Update the internal voice registers
    soundRegsToVoices();

    return 0;
}

/*
    Parse information from the sound registers into the three
    logical voice objects.
*/
void PacmanMachine::soundRegsToVoices( void )
{
    int base = 0;

    for( int i=0; i<3; i++ ) {
        sound_voices_[i].waveform = sound_regs_[ 0x05 + base ] & 0x07;
        sound_voices_[i].volume = sound_regs_[ 0x15 + base ] & 0x0F;

        unsigned f;
        
        f =            (sound_regs_[ 0x14+base ] & 0x0F);
        f = (f << 4) | (sound_regs_[ 0x13+base ] & 0x0F);
        f = (f << 4) | (sound_regs_[ 0x12+base ] & 0x0F);
        f = (f << 4) | (sound_regs_[ 0x11+base ] & 0x0F);
        f = (f << 4);

        if( i == 0 ) { // The first voice has an extra 4-bit of data
            f |= (sound_regs_[ 0x10+base ] & 0x0F); 
        }

        sound_voices_[i].frequency = f;

        base += 5;
    }
}

/*
    Start an input event, telling the emulator for example
    that the joystick has been moved to the down position.
*/
void PacmanMachine::setEventOn( InputEvent event )
{
    switch( event ) {
    case JoystickUp:
        port1_ &= ~0x01;
        break;
    case JoystickLeft:
        port1_ &= ~0x02; 
        break;
    case JoystickRight:
        port1_ &= ~0x04; 
        break;
    case JoystickDown:
        port1_ &= ~0x08; 
        break;
    case CoinInserted:
        port1_ &= ~0x20; 
        break;
    case KeyOnePlayer:
        port2_ &= ~0x20; 
        break;
    case KeyTwoPlayers:
        port2_ &= ~0x40; 
        break;
    }
}

/*
    End an input event, telling the emulator for example
    that the joystick has been released from the down position.
*/
void PacmanMachine::setEventOff( InputEvent event )
{
    switch( event ) {
    case JoystickUp:
        port1_ |= 0x01;
        break;
    case JoystickLeft:
        port1_ |= 0x02;
        break;
    case JoystickRight:
        port1_ |= 0x04;
        break;
    case JoystickDown:
        port1_ |= 0x08;
        break;
    case CoinInserted:
        port1_ |= 0x20;
        break;
    case KeyOnePlayer:
        port2_ |= 0x20;
        break;
    case KeyTwoPlayers:
        port2_ |= 0x40;
        break;
    }
}

void PacmanMachine::decodeCharByte( unsigned char b, unsigned char * charbuf, int charx, int chary, int charwidth )
{
    for( int i=3; i>=0; i-- ) {
        charbuf[charx+(chary+i)*charwidth] = (b & 1) | ((b >> 3) & 2);
        b >>= 1;
    }
}

void PacmanMachine::decodeCharLine( unsigned char * src, unsigned char * charbuf, int charx, int chary, int charwidth )
{
    for( int x=7; x>=0; x-- ) {
        decodeCharByte( *src++, charbuf, x+charx, chary, charwidth );
    }
}

void PacmanMachine::decodeCharSet( unsigned char * mem, unsigned char * charset )
{
    for( int i=0; i<256; i++ ) {
        unsigned char * src = mem + 16*i;
        unsigned char * dst = charset + 64*i;

        decodeCharLine( src,   dst, 0, 4, 8 );
        decodeCharLine( src+8, dst, 0, 0, 8 );
    }
}

void PacmanMachine::decodeSprites( unsigned char * mem, unsigned char * sprite_data )
{
    for( int i=0; i<64; i++ ) {
        unsigned char * src = mem + i*64;
        unsigned char * dst = sprite_data + 256*i;

        decodeCharLine( src   , dst, 8, 12, 16 );
        decodeCharLine( src+ 8, dst, 8,  0, 16 );
        decodeCharLine( src+16, dst, 8,  4, 16 );
        decodeCharLine( src+24, dst, 8,  8, 16 );
        decodeCharLine( src+32, dst, 0, 12, 16 );
        decodeCharLine( src+40, dst, 0,  0, 16 );
        decodeCharLine( src+48, dst, 0,  4, 16 );
        decodeCharLine( src+56, dst, 0,  8, 16 );
    }
}

/*
    Decode one byte from the encoded color palette.

    An encoded palette byte contains RGB information bit-packed as follows:
        
          bit: 7 6 5 4 3 2 1 0
        color: b b g g g r r r
*/
unsigned PacmanMachine::decodePaletteByte( unsigned char value )
{
    unsigned    bit0, bit1, bit2;
    unsigned    red, green, blue;

	bit0 = (value >> 0) & 0x01;
	bit1 = (value >> 1) & 0x01;
	bit2 = (value >> 2) & 0x01;
	red = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

    bit0 = (value >> 3) & 0x01;
	bit1 = (value >> 4) & 0x01;
	bit2 = (value >> 5) & 0x01;
	green = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

    bit0 = 0;
	bit1 = (value >> 6) & 0x01;
	bit2 = (value >> 7) & 0x01;
    blue = 0x21 * bit0 + 0x47 * bit1 + 0x97 * bit2;

    return (blue << 16 ) | (green << 8) | red;
}

/*
    Play and mix the sound voices into the specified buffer.
*/
void PacmanMachine::playSound( int * buf, int len, int samplingRate )
{
    // Clear the buffer
    memset( buf, 0, sizeof(int)*len );

    // To avoid overflow we divide the output frequency by 75, which is
    // the Greatest Common Denominator of 11025 and 96000.
    // Later we use 96000/75 = 1280 (0x500) as the frequency multiplier and this
    // is a safe value because (the max possible frequency) 0xFFFFF * 0x500 does
    // fit in 32 bit.
    // Because 22050 and 44100 are both multiples of 11025, this method allows us
    // to use (faster) integer math and still work without errors at the canonical
    // sampling rates.
    samplingRate /= 75;

    for( int v=0; v<3; v++ ) {
        if( (sound_voices_[v].frequency != 0) && (sound_voices_[v].volume > 0) ) {
            // Ok, the voice is playing
            unsigned offset = sound_voices_[v].wave_offset_;
            int volume = (int) sound_voices_[v].volume;
            int * wave_data = sound_wave_data_ + 32 * sound_voices_[v].waveform;

            for( int i=0; i<len; i++ ) {
                unsigned o = (offset * 1280) / samplingRate;

                buf[i] += wave_data[(o >> 15) & 0x1F] * volume;

                // Note: it is possible to approximate the 96000 Hz frequency of the 
                // chip with the value 88200. This makes it possible to avoid the above
                // multiply and divide altogether because the common sampling rates of
                // 11025, 22050 and 44100 now divide the chip frequency perfectly. So
                // for example if we shift one bit more:
                //   buf[i] += wave_data[(offset >> 14) & 0x1F] * volume;
                // we get the resampled value for 44100 Hz. Similarly, we can get the
                // values for 22050 and 11025 by shifting by 13 and 12 respectively.

                offset = (offset + sound_voices_[v].frequency) & 0xFFFFF;
            }

            sound_voices_[v].wave_offset_ = offset;
        }
    }
}

static unsigned saveBuffer( unsigned char * buf, void * data, unsigned data_size )
{
    memcpy( buf, data, data_size );

    return data_size;
}

static unsigned loadBuffer( void * data, unsigned data_size, unsigned char * buf )
{
    memcpy( data, buf, data_size );

    return data_size;
}

unsigned PacmanMachine::getSizeOfSnapshotBuffer() const
{
    unsigned result = 
        sizeof( ram_ ) +
        sizeof( sound_regs_ ) +
        sizeof( sound_prom_ ) +
        sizeof( video_mem_ ) +
        sizeof( color_mem_ ) +
        sizeof( palette_data_ ) +
        sizeof( color_data_ ) +
        sizeof( charset_rom_ ) +
        sizeof( spriteset_rom_ ) +
        1 + // port1_
        1 + // port2_
        1 + // interrupt_
        1 + // interrupt_enabled_
        1 + // dip_switches_
        8*2 + // sprites_ x
        8*2 + // sprites_ y
        cpu_->getSizeOfSnapshotBuffer();

    return result;
}

static unsigned addUint16( unsigned char * buffer, unsigned u )
{
    *buffer++ = (unsigned char) (u >> 8);
    *buffer++ = (unsigned char) (u);

    return 2;
}

unsigned PacmanMachine::getSnapshot( unsigned char * buffer )
{
    unsigned char * buf = buffer;

    buf += saveBuffer( buf, ram_, sizeof(ram_) );
    buf += saveBuffer( buf, sound_regs_, sizeof(sound_regs_) );
    buf += saveBuffer( buf, sound_prom_, sizeof(sound_prom_) );
    buf += saveBuffer( buf, video_mem_, sizeof(video_mem_) );
    buf += saveBuffer( buf, color_mem_, sizeof(color_mem_) );
    buf += saveBuffer( buf, palette_data_, sizeof(palette_data_) );
    buf += saveBuffer( buf, color_data_, sizeof(color_data_) );
    buf += saveBuffer( buf, charset_rom_, sizeof(charset_rom_) );
    buf += saveBuffer( buf, spriteset_rom_, sizeof(spriteset_rom_) );
    buf += saveBuffer( buf, &port1_, 1 );
    buf += saveBuffer( buf, &port2_, 1 );
    buf += saveBuffer( buf, &interrupt_, 1 );
    buf += saveBuffer( buf, &interrupt_enabled_, 1 );
    buf += saveBuffer( buf, &dip_switches_, 1 );
    for( int i=0; i<=7; i++ ) {
        buf += addUint16( buf, (unsigned) sprites_[i].x );
        buf += addUint16( buf, (unsigned) sprites_[i].y );
    }
    buf += cpu_->getSnapshot( buf );

    return buf - buffer;
}

static unsigned getUint16( unsigned char * buffer )
{
    unsigned result = *buffer++;

    result = (result << 8) | *buffer++;

    return result;
}

void PacmanMachine::restoreSnapshot( unsigned char * buffer )
{
    unsigned char sound_prom_buf[ sizeof(sound_prom_) ];
    unsigned char palette_data_buf[ sizeof(palette_data_) ];
    unsigned char color_data_buf[ sizeof(color_data_) ];

    buffer += loadBuffer( ram_, sizeof(ram_), buffer );
    buffer += loadBuffer( sound_regs_, sizeof(sound_regs_), buffer );
    buffer += loadBuffer( sound_prom_buf, sizeof(sound_prom_buf), buffer );
    buffer += loadBuffer( video_mem_, sizeof(video_mem_), buffer );
    buffer += loadBuffer( color_mem_, sizeof(color_mem_), buffer );
    buffer += loadBuffer( palette_data_buf, sizeof(palette_data_buf), buffer );
    buffer += loadBuffer( color_data_buf, sizeof(color_data_buf), buffer );
    buffer += loadBuffer( charset_rom_, sizeof(charset_rom_), buffer );
    buffer += loadBuffer( spriteset_rom_, sizeof(spriteset_rom_), buffer );
    buffer += loadBuffer( &port1_, 1, buffer );
    buffer += loadBuffer( &port2_, 1, buffer );
    buffer += loadBuffer( &interrupt_, 1, buffer );
    buffer += loadBuffer( &interrupt_enabled_, 1, buffer );
    buffer += loadBuffer( &dip_switches_, 1, buffer );
    for( int i=0; i<=7; i++ ) {
        sprites_[i].x = (int) getUint16( buffer ); buffer += 2;
        sprites_[i].y = (int) getUint16( buffer ); buffer += 2;
        sprites_[i].color = ram_[ 0x4FF0 + i*2 + 1 ];
        sprites_[i].n = ram_[ 0x4FF0 + i*2 ] >> 2;
        sprites_[i].mode = ram_[ 0x4FF0 + i*2 ] & 0x03;
    }
    cpu_->restoreSnapshot( buffer );

    setColorROMs( palette_data_buf, color_data_buf );
    setSoundPROM( sound_prom_buf );
    decodeCharSet( charset_rom_, charmap_ );
    decodeSprites( spriteset_rom_, spritemap_ );
    soundRegsToVoices();
}

void PacmanMachine::drawChar( unsigned char * buffer, int index, int ox, int oy, int color )
{
    buffer += ox + oy*224; // Make the buffer point to the character position
    index *= 64; // Make the index point to the character offset into the character table
    color = (color & 0x3F)*4;

    for( int y=0; y<8; y++ ) {
        for( int x=0; x<8; x++ ) {
            buffer[x] = charmap_[ index++ ] + color;
        }
        buffer += ScreenWidth; // Go to the next line
    }
}

void PacmanMachine::drawSprite( unsigned char * buffer, int index )
{
    PacmanSprite & ps = sprites_[index];

    // Exit now if sprite not visible at all
    if( (ps.color == 0) || (ps.x >= ScreenWidth) || (ps.y < 16) || (ps.y >= (ScreenHeight-32)) ) {
        return;    
    }

    
    // Clip the sprite coordinates to cut the parts that fall off the screen
    int start_x = (ps.x < 0) ? 0 : ps.x;
    int end_x = (ps.x < (ScreenWidth-16)) ? ps.x+16 : ScreenWidth;

    // Prepare variables for drawing
    int color = (ps.color & 0x3F)*4;
    unsigned char * spritemap_base = spritemap_ + ((ps.n & 0x3F)*256);
    
    buffer += ScreenWidth*ps.y;

    // Draw the 16x16 sprite
    for( int y=0; y<16; y++ ) {
        for( int x=start_x; x<end_x; x++ ) {
            int c;
            int o = x-ps.x; // X position relative to the sprite

            if( ps.mode == 0 )                  // Normal
                c = spritemap_base[o+y*16];
            else if( ps.mode == 1 )             // Flip Y
                c = spritemap_base[o+(15-y)*16];
            else if( ps.mode == 2 )             // Flip X
                c = spritemap_base[15-o+y*16];
            else                                // Flip X and Y
                c = spritemap_base[15-o+(15-y)*16];

            if( c ) {
                buffer[x] = c + color;
            }
        }

        buffer += ScreenWidth;
    }
}

/*
    Draw the video into the specified buffer.
*/
void PacmanMachine::renderVideo( unsigned char * buffer )
{
    unsigned char * video = video_mem_;
    unsigned char * color = color_mem_;

    // Draw the background first...
    for( int y=0; y<36; y++ ) {
        for( int x=0; x<28; x++ ) {
            drawChar( buffer, *video, x*8, y*8, *color );
            video++;
            color++;
        }
    }

    // ...then add the sprites
    for( int i=7; i>=0; i-- ) {
        drawSprite( buffer, i );
    }
}
