/*
    Pacman Instructional Emulator
    User interface for Windows

    Copyright (c) 1997-2003,2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <windows.h>
#include <windowsx.h>
#include <mmsystem.h>
#include <ddraw.h>
#include <dsound.h>
#include <stdio.h>
#include <assert.h>

#include "resource.h"

#include "arcade.h"
#include "dib24.h"

unsigned framesPerSecond    = PacmanMachine::VideoFrequency;
unsigned samplesPerSecond   = 44100;    // Valid values: 11025, 22050, 44100
unsigned bytesPerSample     = 2;        // Valid values: 1 or 2
unsigned samplesPerFrame    = samplesPerSecond / framesPerSecond;
unsigned sampleBytesPerFrame= bytesPerSample * samplesPerFrame;

static HWND     hWindow = NULL;         // Handle of main window
static BOOL     boIsActive = FALSE;     // Program active/inactive
static char     szAppName[] = "PacMan";

// Bitmap used for off-screen rendering
static Dib24    dCanvas( PacmanMachine::ScreenWidth, PacmanMachine::ScreenHeight );

// DIP switches
static unsigned char dipDifficulty = PacmanMachine::DipDifficulty_Normal;
static unsigned char dipGhostNames = PacmanMachine::DipGhostNames_Normal;
static unsigned char dipLivesPerGame = PacmanMachine::DipLives_3;   

static char szHelp[] =
    "1 = one player\n"
    "2 = two players\n"
    "3 = insert coin\n"
    "Left/Up/Right/Down = move\n"
    "F5/F6/F7 = save game snapshot\n"
    "Escape = quit";

static char szAbout[] = 
    "PIE - Pacman Instructional Emulator\n"
    "Copyright (c) 1997-2003,2004 Alessandro Scotti\n"
    "\n"
    "http://www.ascotti.org/";

PacmanMachine arcade; // Arcade emulator

LPDIRECTSOUND directSound = 0; // Main direct sound object
IDirectSoundBuffer * directSoundBuf = 0; // Sound buffer
int * rawSoundBuf; // Raw sound buffer used for arcade.playSound()
BOOL applicationInitialized = FALSE;

BOOL showError( char * msg )
{
    MessageBox( 0, msg, "Error", MB_OK | MB_ICONERROR );
    return FALSE;
}

BOOL loadFile( const char * name, unsigned char * buf, int len, BOOL showErrorMsg = FALSE )
{
    char msg[200];
    FILE * f = fopen( name, "rb" );

    if( f == NULL ) {
        sprintf( msg, "Cannot open '%s'", name );
        if( showErrorMsg ) showError( msg );
        return FALSE;
    }
    
    int n = fread( buf, len, 1, f );
    
    fclose( f );

    if( n != 1 ) {
        sprintf( msg, "Error reading from '%s'", name );
        if( showErrorMsg ) showError( msg );
        return FALSE;
    }

    return TRUE;
}

int menuCheck( int condition )
{
    return condition ? MF_CHECKED : MF_UNCHECKED;
}

/*
    Sets the game DIP switches and updates the program menu accordingly.
*/
void setDipSwitches( BOOL reset = TRUE )
{
    HMENU   hMenu = GetMenu( hWindow );

    CheckMenuItem( hMenu, IDM_NORMAL, menuCheck(dipDifficulty == PacmanMachine::DipDifficulty_Normal) );
    CheckMenuItem( hMenu, IDM_HARD, menuCheck(dipDifficulty == PacmanMachine::DipDifficulty_Hard) );
  
    CheckMenuItem( hMenu, IDM_LIFE_0, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_1) );
    CheckMenuItem( hMenu, IDM_LIFE_1, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_2) );
    CheckMenuItem( hMenu, IDM_LIFE_2, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_3) );
    CheckMenuItem( hMenu, IDM_LIFE_3, menuCheck(dipLivesPerGame == PacmanMachine::DipLives_5) );

    CheckMenuItem( hMenu, IDM_GHOST_0, menuCheck(dipGhostNames == PacmanMachine::DipGhostNames_Normal) );
    CheckMenuItem( hMenu, IDM_GHOST_1, menuCheck(dipGhostNames == PacmanMachine::DipGhostNames_Alternate) );

    arcade.setDipSwitches( PacmanMachine::DipPlay_OneCoinOneGame | PacmanMachine::DipBonus_10000 |
        dipDifficulty | dipGhostNames | dipLivesPerGame );

    if( reset ) {
        arcade.reset();
    }
}

/*
    Saves a snapshot of the game.
*/
void takeSnapshot( char * filename )
{
    unsigned bufsize = arcade.getSizeOfSnapshotBuffer();
    unsigned char * buffer = new unsigned char [ bufsize ];

    arcade.getSnapshot( buffer );

    FILE * f = fopen( filename, "wb" );

    assert( f != NULL );

    fwrite( buffer, 1, bufsize, f );
    fclose( f );

    delete buffer;
}

/*
    Restores a snapshot of the game.
*/
void restoreSnapshot( char * filename )
{
    unsigned bufsize = arcade.getSizeOfSnapshotBuffer();
    unsigned char * buffer = new unsigned char [ bufsize ];

    FILE * f = fopen( filename, "rb" );

    assert( f != NULL );

    assert( fread( buffer, 1, bufsize, f ) == bufsize );

    fclose( f );

    arcade.restoreSnapshot( buffer );

    delete buffer;

    dipDifficulty = arcade.getDipSwitches() & PacmanMachine::DipDifficulty_Mask;
    dipGhostNames = arcade.getDipSwitches() & PacmanMachine::DipGhostNames_Mask;
    dipLivesPerGame = arcade.getDipSwitches() & PacmanMachine::DipLives_Mask;
    setDipSwitches( FALSE );
}

/*
    Initializes the sound library and creates the DirectX sound objects
    used by the application.
*/
BOOL initSound( void )
{
    BOOL result = FALSE;

    if( DS_OK == DirectSoundCreate(NULL, &directSound, NULL) ) {
        directSound->SetCooperativeLevel( hWindow, DSSCL_NORMAL );

        DSBUFFERDESC    bufdesc = { 0 };
        WAVEFORMATEX    wfx = { 0 };

        wfx.wFormatTag = WAVE_FORMAT_PCM;
        wfx.nChannels = 1;
        wfx.nSamplesPerSec = samplesPerSecond;
        wfx.wBitsPerSample = bytesPerSample * 8;
        wfx.nBlockAlign = (wfx.nChannels * wfx.wBitsPerSample) / 8;
        wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;

        bufdesc.dwSize = sizeof(bufdesc);
        bufdesc.dwFlags = DSBCAPS_STICKYFOCUS;
        bufdesc.dwBufferBytes = wfx.nBlockAlign * samplesPerFrame * 2;
        bufdesc.lpwfxFormat = &wfx;

        if( SUCCEEDED(directSound->CreateSoundBuffer( &bufdesc, &directSoundBuf, NULL)) ) {
            LPVOID  p1, p2;
            DWORD   size1, size2;

            directSoundBuf->GetFormat( &wfx, sizeof(wfx), 0 );

            assert( samplesPerSecond == wfx.nSamplesPerSec );

            if( SUCCEEDED(directSoundBuf->Lock(0, bufdesc.dwBufferBytes, &p1, &size1, &p2, &size2, 0)) ) {
                ZeroMemory( p1, size1 );

                if( size2 != 0 ) 
                    ZeroMemory( p2, size2 );

                directSoundBuf->Unlock( p1, size1, p2, size2 );
                
                result = TRUE;
            }
            else {
                showError( "Error: cannot lock the DirectSoundBuffer object!" );
                directSoundBuf->Release();
                directSoundBuf = 0;
            }
        }
        else {
            showError( "Error: cannot create DirectSoundBuffer object!" );
            directSoundBuf = 0;
        }
    }
    else {
        showError( "Error: cannot create DirectSound object!" );
        directSound = 0;
    }

    return result;
}

/*
    Activates/deactivates the program.

    Note: when the program is not active, emulation and sound playing are stopped.
*/
void setActive( BOOL active )
{
    if( ! applicationInitialized )
        return;

    active = active ? TRUE : FALSE; // Normalize the boolean values!

    if( active != boIsActive ) {
        boIsActive = active;
        if( active )
            directSoundBuf->Play( 0, 0, DSBPLAY_LOOPING );
        else
            directSoundBuf->Stop();
    }
}

/*
    Initializes the game engine.
*/
BOOL gameInit( void )
{
    BOOL romLoaded = FALSE;

    // Initialize the sound library
    if( ! initSound() )
        return FALSE;

    rawSoundBuf = new int[ samplesPerFrame ];

    // Load game ROMs
    unsigned char   buffer[ 0x6000 ];
    unsigned char   palette[0x20];
    unsigned char   color[0x100];

    if( (! loadFile( "82s126.4a", color, sizeof(color), TRUE)) ||
        (! loadFile( "82s123.7f", palette, sizeof(palette), TRUE)) )
    {
        return FALSE;
    }

    romLoaded = loadFile( "pacman.rom", buffer, sizeof(buffer));

    if( ! romLoaded ) {
        // ROM not found, try the split Midway ROMs
        romLoaded = loadFile( "namcopac.6e", buffer, 0x1000 ) &&
            loadFile( "namcopac.6f", buffer+0x1000, 0x1000 ) &&
            loadFile( "namcopac.6h", buffer+0x2000, 0x1000 ) &&
            loadFile( "namcopac.6j", buffer+0x3000, 0x1000 ) &&
            loadFile( "pacman.5e", buffer+0x4000, 0x1000 ) &&
            loadFile( "pacman.5f", buffer+0x5000, 0x1000 );
    }

    if( ! romLoaded ) {
        // ROM not found, try the split Namco ROMs
        romLoaded = loadFile( "pacman.6e", buffer, 0x1000 ) &&
            loadFile( "pacman.6f", buffer+0x1000, 0x1000 ) &&
            loadFile( "pacman.6h", buffer+0x2000, 0x1000 ) &&
            loadFile( "pacman.6j", buffer+0x3000, 0x1000 ) &&
            loadFile( "pacman.5e", buffer+0x4000, 0x1000 ) &&
            loadFile( "pacman.5f", buffer+0x5000, 0x1000 );
    }

    if( ! romLoaded ) {
        showError( "Game ROMs not found!" );
        return FALSE;
    }

    // Initialize the emulator
    arcade.setROM( buffer );
    arcade.setVideoROMs( buffer+0x4000, buffer+0x5000 );
    arcade.setColorROMs( palette, color );
    arcade.reset();

    applicationInitialized = TRUE;

    return TRUE;
}

/*
    Runs the game engine for one frame.
*/
void CALLBACK gameProc( void )
{
    static unsigned bufferIndex = 0;

    if( ! boIsActive )
        return;

    int i;

    // Run the machine for a frame
    arcade.run();

    // Draw the video image into the off-screen bitmap
    if( bufferIndex == 0 ) {
        unsigned char video_buffer[ PacmanMachine::ScreenWidth * PacmanMachine::ScreenHeight ];
        unsigned char * vbuf = video_buffer;
        const unsigned * palette = arcade.getPalette();
        
        arcade.renderVideo( vbuf );

        for( int y=0; y<PacmanMachine::ScreenHeight; y++ ) {
            for( int x=0; x<PacmanMachine::ScreenWidth; x++ ) {
                dCanvas.setFastPixels( x, y, palette[ *vbuf++ ] );
            }
        }
    }

    // Prepare the sound buffer
    arcade.playSound( rawSoundBuf, samplesPerFrame, samplesPerSecond );

    // Now wait until the sound buffer for the last frame has finished playing:
    // this delay also allows us to keep the frames going at the correct rate
    DWORD dwPlayCursor;
    DWORD dwWriteCursor;

    if( bufferIndex ) {
        do {
            directSoundBuf->GetCurrentPosition( &dwPlayCursor, &dwWriteCursor );
        } while( dwPlayCursor < sampleBytesPerFrame );
    }
    else {
        do {
            directSoundBuf->GetCurrentPosition( &dwPlayCursor, &dwWriteCursor );
        } while( dwPlayCursor >= sampleBytesPerFrame );
    }

    // Update the video
    if( bufferIndex == 0 ) {
        HDC dc = GetDC( hWindow );
        dCanvas.stretch( dc, 0, 0, PacmanMachine::ScreenWidth, PacmanMachine::ScreenHeight );
        ReleaseDC( hWindow, dc );
    }

    // Update the sound
    void *  pwBuf = 0;
    DWORD   dwBufSize = 0;

    if( DS_OK == directSoundBuf->Lock( bufferIndex*sampleBytesPerFrame, 
        sampleBytesPerFrame, 
        &pwBuf, 
        &dwBufSize, 
        0, 0, 0 ) )
    {
        assert( pwBuf != 0 );
        assert( dwBufSize == sampleBytesPerFrame );

        // Write into the sound buffer
        int * rawBuf = rawSoundBuf;
        BYTE * sndBuf = (BYTE *)pwBuf;

        if( bytesPerSample == 1 ) {
            for( i=0; i<(int)samplesPerFrame; i++ ) {
                *sndBuf++ = (BYTE) (*rawBuf++ / 3) + 0x80;
            }
        }
        else {
            for( i=0; i<(int)samplesPerFrame; i++ ) {
                WORD w = (WORD)(*rawBuf++ * 85);
                *sndBuf++ = (BYTE) (w);
                *sndBuf++ = (BYTE) (w >> 8);
            }
        }

        directSoundBuf->Unlock( pwBuf, dwBufSize, 0, 0 );
    }

    // Flip the buffer!
    bufferIndex = 1 - bufferIndex;
}

/*
    Terminates the game engine.
*/
void gameTerm()
{
    if( directSound ) {
        if( directSoundBuf ) {
            directSoundBuf->Stop();
            directSoundBuf->Release();
        }
        directSound->Release();
        directSound;
    }

    delete rawSoundBuf;
}

/*
    Window procedure.
*/
LRESULT PASCAL WndProc( HWND hWnd, UINT wMsg, WPARAM wParam, LPARAM lParam )
{
    PAINTSTRUCT stPS;
    HDC hDC;
    static BOOL boSavedActiveFlag = FALSE;

    switch( wMsg ) {
        // No need to erase the background
        case WM_ERASEBKGND:
            return TRUE;
        // Paint the window by just blitting the game video
        case WM_PAINT:
            hDC = BeginPaint( hWnd, &stPS );
            dCanvas.stretch( hDC, 0, 0, PacmanMachine::ScreenWidth, PacmanMachine::ScreenHeight );
            EndPaint( hWnd, &stPS );
            return TRUE;
        // Application is activating/deactivating
        case WM_ACTIVATEAPP:
            setActive( (BOOL)wParam );
            break;
        case WM_ENTERMENULOOP:
            boSavedActiveFlag = boIsActive;
            setActive( FALSE );
            break;
        case WM_EXITMENULOOP:
            setActive( boSavedActiveFlag );
            break;
        // Menu commands
        case WM_COMMAND:
            switch( LOWORD(wParam) ) {
            case IDM_NEW:
                arcade.reset();
                break;
            case IDM_SNAP_0:
                restoreSnapshot( "snap.f5" );
                break;
            case IDM_SNAP_1:
                restoreSnapshot( "snap.f6" );
                break;
            case IDM_SNAP_2:
                restoreSnapshot( "snap.f7" );
                break;
            case IDM_EXIT:
                PostMessage( hWnd, WM_CLOSE, 0, 0 );
                break;
            case IDM_HELP_KEYS:
                MessageBox( 0, szHelp, "Keys", MB_OK | MB_ICONINFORMATION );
                break;
            case IDM_ABOUT:
                MessageBox( 0, szAbout, szAppName, MB_OK | MB_ICONINFORMATION );
                break;
            case IDM_LIFE_0:
                dipLivesPerGame = PacmanMachine::DipLives_1;
                setDipSwitches();
                break;
            case IDM_LIFE_1:
                dipLivesPerGame = PacmanMachine::DipLives_2;
                setDipSwitches();
                break;
            case IDM_LIFE_2:
                dipLivesPerGame = PacmanMachine::DipLives_3;
                setDipSwitches();
                break;
            case IDM_LIFE_3:
                dipLivesPerGame = PacmanMachine::DipLives_5;
                setDipSwitches();
                break;
            case IDM_HARD:
                dipDifficulty = PacmanMachine::DipDifficulty_Hard;
                setDipSwitches();
                break;
            case IDM_NORMAL:
                dipDifficulty = PacmanMachine::DipDifficulty_Normal;
                setDipSwitches();
                break;
            case IDM_GHOST_0:
                dipGhostNames = PacmanMachine::DipGhostNames_Normal;
                setDipSwitches();
                break;
            case IDM_GHOST_1:
                dipGhostNames = PacmanMachine::DipGhostNames_Alternate;
                setDipSwitches();
                break;
            }
            break;
        // Keyboard events
        case WM_KEYDOWN:
            switch( wParam ) {
            case ' ':
                setActive( ! boIsActive );
                break;
            case '1':
                arcade.setEventOn( PacmanMachine::KeyOnePlayer );
                break;
            case '2':
                arcade.setEventOn( PacmanMachine::KeyTwoPlayers );
                break;
            case '3':
                arcade.setEventOn( PacmanMachine::CoinInserted );
                break;
            case VK_LEFT: 
                arcade.setEventOn( PacmanMachine::JoystickLeft );
                break;
            case VK_RIGHT:
                arcade.setEventOn( PacmanMachine::JoystickRight );
                break;
            case VK_UP:
                arcade.setEventOn( PacmanMachine::JoystickUp );
                break;
            case VK_DOWN:
                arcade.setEventOn( PacmanMachine::JoystickDown );
                break;
            case VK_ESCAPE:
                PostMessage( hWnd, WM_CLOSE, 0, 0 );
                return 0;
            }
            break;
        case WM_KEYUP:
            switch( wParam ) {
            case '1':
                arcade.setEventOff( PacmanMachine::KeyOnePlayer );
                break;
            case '2':
                arcade.setEventOff( PacmanMachine::KeyTwoPlayers );
                break;
            case '3':
                arcade.setEventOff( PacmanMachine::CoinInserted );
                break;
            case VK_LEFT: 
                arcade.setEventOff( PacmanMachine::JoystickLeft );
                break;
            case VK_RIGHT:
                arcade.setEventOff( PacmanMachine::JoystickRight );
                break;
            case VK_UP:
                arcade.setEventOff( PacmanMachine::JoystickUp );
                break;
            case VK_DOWN:
                arcade.setEventOff( PacmanMachine::JoystickDown );
                break;
            case VK_F4:
                {
                    // Cheat: add a life
                    unsigned char * ram = (unsigned char *) arcade.getRAM();
                    ram[0x4E14]++;
                    ram[0x4E15]++;
                }
            case VK_F5:
                takeSnapshot( "snap.f5" );
                break;
            case VK_F6:
                takeSnapshot( "snap.f6" );
                break;
            case VK_F7:
                takeSnapshot( "snap.f7" );
                break;
            case VK_F9:
                restoreSnapshot( "snap.f5" );
                break;
            }
            break;
        // Destroy window
        case WM_DESTROY:
            PostQuitMessage( 0 );
            break;
    }
    return DefWindowProc( hWnd, wMsg, wParam, lParam );
}

/*
    Application main.
*/
int WINAPI WinMain( HINSTANCE hInstance, HINSTANCE hPrevInst, LPSTR lpCmdLine, int nCmdShow )
{
    MSG         stMsg;
    WNDCLASS    stWndClass;
    HWND        hWnd;

    /* Setup window class structure */
    stWndClass.style = CS_DBLCLKS;
    stWndClass.lpfnWndProc = WndProc;
    stWndClass.cbClsExtra = 0;
    stWndClass.cbWndExtra = 0;
    stWndClass.hInstance = hInstance;
    stWndClass.hIcon = LoadIcon( hInstance, MAKEINTRESOURCE(IDI_MAIN) );
    stWndClass.hCursor = LoadCursor( NULL, IDC_ARROW );
    stWndClass.hbrBackground = (HBRUSH)GetStockObject( BLACK_BRUSH );
    stWndClass.lpszMenuName = MAKEINTRESOURCE( IDM_MAIN );
    stWndClass.lpszClassName = "PacmanDemoProgram";

    /* Register window class */
    if( ! RegisterClass( &stWndClass ) )
        return 0;

    DWORD   wndStyle = WS_VISIBLE | WS_OVERLAPPEDWINDOW;
    RECT    rc;

    rc.left = 0; 
    rc.top = 0; 
    rc.right = PacmanMachine::ScreenWidth; 
    rc.bottom = PacmanMachine::ScreenHeight;

    AdjustWindowRect( &rc, wndStyle, TRUE );

    /* Create window */
    hWnd = CreateWindowEx( 0,               // Extended style
        stWndClass.lpszClassName,           // Class
        szAppName,                          // Title
        wndStyle,                           // Style
        CW_USEDEFAULT,                      // X
        CW_USEDEFAULT,                      // Y
        rc.right-rc.left, 
        rc.bottom-rc.top,
        NULL,                               // Parent or owner window
        NULL,                               // Menu or child window id
        hInstance,                          // Instance handle
        NULL );                             // Pointer to window creation data

    if( hWnd == NULL )
        return 0;

    hWindow = hWnd;

    /* Main loop */
    if( gameInit() ) {
        ShowWindow( hWnd, SW_SHOWNORMAL );
        UpdateWindow( hWnd );

        setActive( TRUE );

        while( TRUE ) {
            if( PeekMessage( &stMsg, NULL, 0, 0, PM_NOREMOVE ) ) {
                if( !GetMessage( &stMsg, NULL, 0, 0 ) )
                    break;
                TranslateMessage( &stMsg );
                DispatchMessage( &stMsg );
            }
            else if( boIsActive ) {
                gameProc();
            }
            else {
                WaitMessage();
            }
        }

        gameTerm();
    }

    return 0;
}
