/*
    Z80 emulator

    Copyright (c) 1996-2003,2004 Alessandro Scotti
    http://www.ascotti.org/

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "z80.h"

Z80::Z80( Z80Environment & env )
    : env_( env )
{
    reset();
}

Z80::Z80( const Z80 & cpu )
    : env_( cpu.env_ )
{
    operator = ( cpu );
}

Z80 & Z80::operator = ( const Z80 & cpu )
{
    B = cpu.B;
    C = cpu.C;
    D = cpu.D;
    E = cpu.E;
    H = cpu.H;
    L = cpu.L;
    A = cpu.A;
    F = cpu.F;
    B1 = cpu.B1;
    C1 = cpu.C1;
    D1 = cpu.D1;
    E1 = cpu.E1;
    H1 = cpu.H1;
    L1 = cpu.L1;
    A1 = cpu.A1;
    F1 = cpu.F1;
    IX = cpu.IX;
    IY = cpu.IY;
    PC = cpu.PC;
    SP = cpu.SP;
    I = cpu.I;
    R = cpu.R;

    iflags_ = iflags_;
    cycles_ = cycles_;

    return *this;
}

void Z80::reset()
{
    B = B1 = 0; 
    C = C1 = 0;
    D = D1 = 0; 
    E = E1 = 0;
    H = H1 = 0;
    L = L1 = 0;
    A = A1 = 0;
    F = F1 = 0;
    IX = 0;
    IY = 0;
    PC = 0;
    SP = 0xF000;
    I = 0;
    R = 0;

    iflags_ = 0;
    cycles_ = 0;
}

unsigned Z80::getSizeOfSnapshotBuffer() const
{
    unsigned result =
        8*2 +   // 8-bit registers
        1 +     // I
        1 +     // R
        2 +     // IX
        2 +     // IY
        2 +     // PC
        2 +     // SP
        4 +     // iflags_
        4;      // cycles_    

    return result;
}

static unsigned addUint16( unsigned char * buffer, unsigned u )
{
    *buffer++ = (unsigned char) (u >> 8);
    *buffer++ = (unsigned char) (u);

    return 2;
}

unsigned Z80::getSnapshot( unsigned char * buffer )
{
    unsigned char * buf = buffer;

    *buf++ = A; *buf++ = A1;
    *buf++ = B; *buf++ = B1;
    *buf++ = C; *buf++ = C1;
    *buf++ = D; *buf++ = D1;
    *buf++ = E; *buf++ = E1;
    *buf++ = H; *buf++ = H1;
    *buf++ = L; *buf++ = L1;
    *buf++ = F; *buf++ = F1;

    *buf++ = I;
    *buf++ = R;

    buf += addUint16( buf, IX );
    buf += addUint16( buf, IY );
    buf += addUint16( buf, PC );
    buf += addUint16( buf, SP );

    buf += addUint16( buf, iflags_ >> 16 );
    buf += addUint16( buf, iflags_ );
    buf += addUint16( buf, cycles_ >> 16 );
    buf += addUint16( buf, cycles_ );

    return buffer - buf;
}

static unsigned getUint16( unsigned char * buffer )
{
    unsigned result = *buffer++;

    result = (result << 8) | *buffer++;

    return result;
}

void Z80::restoreSnapshot( unsigned char * buffer )
{
    unsigned char * buf = buffer;

    A = *buf++; A1 = *buf++;
    B = *buf++; B1 = *buf++;
    C = *buf++; C1 = *buf++;
    D = *buf++; D1 = *buf++;
    E = *buf++; E1 = *buf++;
    H = *buf++; H1 = *buf++;
    L = *buf++; L1 = *buf++;
    F = *buf++; F1 = *buf++;

    I = *buf++;
    R = *buf++;

    IX = getUint16( buf ); buf += 2;
    IY = getUint16( buf ); buf += 2;
    PC = getUint16( buf ); buf += 2;
    SP = getUint16( buf ); buf += 2;

    iflags_ = getUint16( buf );                 buf += 2;
    iflags_ = (iflags_ << 16) | getUint16(buf); buf += 2;
    cycles_ = getUint16( buf );                 buf += 2;
    cycles_ = (cycles_ << 16) | getUint16(buf); buf += 2;
}

void Z80::step()
{
    if( iflags_ & Halted ) {
        cycles_ += 4;
        return;
    }

    unsigned op = fetchByte();

    // Execute
    cycles_ += OpInfo_[ op ].cycles;
    (this->*(OpInfo_[ op ].handler))();

    R = (R+1) & 0x7F;

    PC &= 0xFFFF;
    SP &= 0xFFFF;
}

void Z80::interrupt( unsigned char data )
{
    if( iflags_ & IFF1 ) {
        if( iflags_ & Halted ) {
            iflags_ &= ~Halted;
            PC++;
        }

        switch( getInterruptMode() ) {
        case 0:
            (this->*(OpInfo_[ data ].handler))();
            cycles_ += 11;
            break;
        case 1:
            callSub( 0x38 );
            cycles_ += 11;
            break;
        case 2:
            iflags_ &= ~(IFF1 | IFF2);
            callSub( env_.readWord( ((unsigned)I) << 8 | (data & 0xFE) ) );
            cycles_ += 19;
            break;
        }
    }
}

void Z80::nmi()
{
    iflags_ &= ~IFF1;   // Disable maskable interrupts

    if( iflags_ & Halted ) {
        iflags_ &= ~Halted;
        PC++;
    }

    callSub( 0x66 );

    cycles_ += 11;
}
